#ifndef _PHOTOMATIX_IMAGE_H
#define _PHOTOMATIX_IMAGE_H

#ifdef __INTBUILD
#include "../Features.h"
#else
#include "Features.h"
#endif

#ifndef OLD_CPLUSPLUS
#ifdef _WIN32
#define OLD_CPLUSPLUS (_MSC_VER < 1900)
#else
#define OLD_CPLUSPLUS (__cplusplus <= 199711L)
#endif
#if OLD_CPLUSPLUS
#define noexcept throw()
#endif
#endif

#if defined(_WIN32) && !defined(SWIG)
#define EXPORT __declspec(dllexport)
#else
#define EXPORT
#endif

/// \file Image.h
/// \brief Contains the definition for the Image class, which stores both source and processed images.
///

namespace photomatix {

	class PhotomatixEngine;

	/// <summary>
	/// Color profile to use when reading RAW images.
	/// </summary>
	enum class EColorProfile
	{
		sRgb = 0,
		AdobeRgb = 1,
		ProPhotoRgb = 2
	};

	/// <summary>
	/// Orientation of loaded and saved images, obtained by Image::getOrientation().
	/// </summary>
	enum class EImageOrientation
	{
		normal = 1,
		flippedHorizontally = 2,
		upsideDown = 3,
		flippedVertically = 4,
		flippedHorizRot270Deg = 5,
		rotated90Clockwise = 6,
		flippedVertRot90Deg = 7,
		rotated270Clockwise = 8
	};

	
	/// <summary>
	/// Stores a source or result image.
	/// </summary>
	/// You never directly create an Image.  They are always created by a PhotomatixEngine object.
	/// For result images, you can get the 32-bit HDR Image merged from source images by calling PhotomatixEngine::getHDRImage, and a merged Image that has been processed 
	/// with a preset from PhotomatixEngine::processWithPreset.  You can then enhance and resize these images, then save them, or use their pixel data directly.
	/// 
	/// Source Images are created by adding them to a PhotomatixEngine object, using PhotomatixEngine::addImageFromFile or PhotomatixEngine::addImageFromData.
	/// These member functions return pointers to the Image objects that they add to the source collection.  It is possible to enhance or resize these also, before
	/// they are merged to HDR.  Source images for merging can be loaded from JPEG, TIFF (8 and 16 bit), and DNG (8 and 16 bit) formats, as well as RAW images from
	/// many cameras.  32-bit HDR images for tone mapping can be loaded from OpenEXR, Radiance, floating-point TIFF and DNG formats.
	/// 
	/// You can get pixel image data for Images using the getData member function.  This gives access to the internal image data for the object.  For 8- and 16-bit images,
	/// image data is stored with interleaved channel values per pixel.  This means, for example, for an image that has red, green and blue channels, that the pixel data is stored like this:
	/// RGBRGBRGBRGB...
	/// Each row will have the number of pixels corresponding to the width of the images, obtained by calling getWidth.  There may be additional padding bytes after the end of the row.
	/// You can get the total number of pixels in each row by calling the getRowStride member function.
	/// 
	/// 32-bit images have data stored in RGBE format, which is documented ***here***.  This is a compact format, but not so easy to work with.  You can export the data to
	/// interleaved channel format with floating-point data for each pixel, using the getFloatData() member function.
	/// 
	/// The example source file @ref ImageData.cpp shows in detail how to access and process the data for all image bit depths.
	class EXPORT Image
	{
	public:	

		///@{
		/// @name Basic Properties
		
		/// <summary>
		/// Width of the image, in pixels
		/// </summary>
		int getWidth() const noexcept;
		
		/// <summary>
		/// Height of the image, in pixels
		/// </summary>
		int getHeight() const noexcept;

		/// <summary>
		/// Bits per sample of the image
		/// </summary>
		int getBps() const noexcept;
		
		/// <summary>
		/// Samples per pixel of the image
		/// </summary>
		int getSpp() const noexcept;

		/// <summary>
		/// Was the image loaded from a RAW file
		/// </summary>
		bool isRAW() const noexcept;

		/// <summary>
		/// Was the image loaded from a RAW file (including RAWs in DNG format), or a non-RAW DNG?
		/// </summary>
		bool isRAWOrDNG() const noexcept;

		/// <summary>
		/// Returns the file name of the image, if it was loaded from a file, or the empty string otherwise.
		/// </summary>
		const char* getFileName() const noexcept;

		///@}

#ifdef POSTPROCESSING_INCLUDED
		///@{
		/// @name Postprocessing Actions

		/// <summary>
		/// Apply sharpening to image in place.
		/// </summary>
		/// <param name="amount">Amount of sharpening: must be greater than 0</param>
		/// <param name="radius">Sharpening radius: valid values are between 0.5 and 3</param>
		/// <param name="threshold">Sharpening threshold: valid values are between 0 and 20</param>
		void sharpen(float amount, float radius, float threshold);

		/// <summary>
		/// Apply contrast selectively to the highlights, lights, shadows and darks of the image in place.
		/// </summary>
		/// <param name="contrastAdjustments">An array of contrast adjustments for highlights, lights, shadows and darks in that order, with values between 0 (no change) and 100 (maximum change).</param>
		void addContrast(float contrastAdjustments[4]);

		/// <summary>
		/// Apply sharpening and selective contrast to the image in place.
		/// </summary>
		/// This performs the same function as calling sharpen() and addContrast(), but more efficiently then calling one and
		/// then the other.  Use this if you want to apply both sharpening and selective contrast.
		/// <param name="sharpenAmount">Amount of sharpening: must be greater than 0</param>
		/// <param name="sharpenRadius">Sharpening radius: valid values are between 0.5 and 3</param>
		/// <param name="sharpenThreshold">Sharpening threshold: valid values are between 0 and 20</param>
		/// <param name="contrastAdjustments">Contrast adjustments for highlights, lights, shadows and darks in that order.</param>
		void addContrastAndSharpen(float sharpenAmount, float sharpenRadius, float sharpenThreshold, float contrastAdjustments[4]);

		/// <summary>
		/// Resize image *in place* to the specified dimensions.  If either width or height is -1, it will be calculated from the other parameter, based on the aspect ratio of the original image.
		/// </summary>
		/// <param name="width">Width to resize, or -1</param>
		/// <param name="height">Height to resize, or -1</param>
		void resize(int width, int height);

		///@}
#endif


		///@{
		///@name Saving and Loading

		/// <summary>
		/// Write the image as a JPEG.
		/// NOTE: Will overwrite an existing file with the same name.
		/// </summary>
		/// <param name="fileName">Output file name</param>
		/// <param name="quality">JPEG quality to save with, from 0 (lowest quality, smallest file size) to 100 (highest quality, largest file size)</param>
		void saveAsJpeg(const char* fileName, int quality) const;

		/// <summary>
		/// Write the image as TIFF.
		/// NOTE: Will overwrite an existing file with the same name.
		/// </summary>
		/// <param name="fileName">Output file name</param>
		/// <param name="bitDepth">Bit depth to save the image with.  Must be 8 or 16 for an LDR image, and 16 or 32 for an HDR image.  Defaults to 8 for LDR, 32 for HDR.  For an HDR image, a bit depth of 16 will save as half float, and 32 as full float.</param>
		/// <param name="normalise">Used only for HDR images.  If true, scale the pixel values to a fixed range.  This can be helpful to avoid scenes appearing too bright or dark when viewing them with Lightroom's default exposure.  Leave as false to keep the unscaled values.</param>
		void saveAsTiff(const char* fileName, int bitDepth = 0, bool normalise = false) const;

#if defined(_WIN32) || defined(__macos)
		/// <summary>
		/// Write the images as DNG.
		/// NOTE: Will overwrite an existing file with the same name.
		/// </summary>
		/// <param name="fileName">Output file name</param>
		/// <param name="bitDepth">Bit depth to save the image with.  Must be 8 or 16 for an LDR image, and 16 or 32 for an HDR image.  Defaults to 8 for LDR, 32 for HDR.  For an HDR image, a bit depth of 16 will save as half float, and 32 as full float.</param>
		/// Bit depth must be 8 or 16 for an LDR image, and 16 or 32 for an HDR image.  Defaults to 8 for LDR, 32 for HDR.
		/// For an HDR image, a bit depth of 16 will save as half float, and 32 as full float.
		/// This is not available on Linux systems.
		void saveAsDNG(const char* fileName, int bitDepth = 0) const;
#endif

		/// <summary>
		/// Write a 32-bit image in OpenEXR format.  Valid only for 32-bit images.
		/// NOTE: Will overwrite an existing file with the same name.
		/// </summary>
		/// <param name="fileName">Output file name</param>
		void saveAsEXR(const char* fileName) const;

		/// <summary>
		/// Write a 32-bit image in Radiance (HDR) format.  Valid only for 32-bit images.
		/// NOTE: Will overwrite an existing file with the same name.
		/// </summary>
		/// <param name="fileName">Output file name</param>
		void saveAsRadiance(const char* fileName) const;

		/// <summary>
		/// Has the image data been loaded.  If false, only the metadata has been loaded.
		/// </summary>
		bool isLoaded() const noexcept;

		/// <summary>
		/// Load the data of the image.  This is normally called automatically when generating the 32-bit HDR image, or processing with
		/// an HDRPreset, but can be called manually to access source pixel data using getData() before merging to HDR.
		/// Throws a PhotomatixException if image couldn't be read, or std::bad_alloc if there is insufficent memory.
		/// </summary>
		/// If loading a RAW image, this function takes into account the requested color temperature and color profile, if it was provided by calling
		/// setRAWColorTemperature() and setRAWColorProfile().  
		void load();


		///@}

		///@{
		///@name Accessing Image Pixel Data

		/// <summary>
		/// Access the raw data bytes for the pixel data of the image.
		/// </summary>
		/// For 8- and 16-bit images, the data is represented as getSpp() samples per pixel, interleaved, with getWidth() pixels per row.  Rows may be padded.  Use getRowStride() or
		/// getRowStrideBytes() to determine the length of each row.  For 8-bit images, each sample is an unsigned byte, whereas for 16-bit images, each sample is an unsigned short.
		/// 32-bit images are stored in RGBE format with 4 bytes per pixel.  Use getFloatData() to save a copy of the 32-bit data as getSpp() float samples per pixel, interleaved.
		/// Note that this is a pointer to data that is deallocated when its Image object is destructed.  Make a copy of it if you need to store it longer than its Image object.
		unsigned char* getData() const noexcept;

		/// <summary>
		/// Access the raw data bytes for the pixel data of the image.
		/// </summary>
		/// For 8- and 16-bit images, the data is represented as getSpp() samples per pixel, interleaved, with getWidth() pixels per row.  Rows may be padded.  Use getRowStride() or
		/// getRowStrideBytes() to determine the length of each row.  For 8-bit images, each sample is an unsigned byte, whereas for 16-bit images, each sample is an unsigned short.
		/// 32-bit images are stored in RGBE format with 4 bytes per pixel.  Use getFloatData() to save a copy of the 32-bit data as getSpp() float samples per pixel, interleaved.
		/// Note that this is a pointer to data that is deallocated when its Image object is destructed.  Make a copy of it if you need to store it longer than its Image object.
		unsigned char* getData() noexcept;

		/// <summary>
		/// The size in bytes of the data for the image, pointed to by getData()
		/// </summary>
		size_t getDataLength() const noexcept;

		/// <summary>
		/// Save a copy of the image data to outData, as float samples, getSpp() samples per pixel.  NOTE: This is valid only for 32-bit images.  An exception will be thrown if this is called on images with other bit depths.
		/// </summary>
		/// <param name="outData">Where to write the data.  Must be at least getFloatDataSize() bytes in size.</param>
		/// <param name="rowStride">The number of samples (not bytes) per row to use when writing data.  You can omit this parameter to use the default row stride of getSpp() * getWidth().  
		/// If a value less than this is passed in, the default row stride will be used instead.  If it is greater than the default row stride, then each row will have uninitialised padding data following the copied data.</param>
		void getFloatData(float* outData, size_t rowStride = 0) const;

		/// <summary>
		/// Returns the size in bytes of the buffer required for getFloatData() to save floating point data from the image, for 32-bit images.
		/// NOTE: This is valid only for 32-bit images.  An exception will be thrown if this is called on images with other bit depths.
		/// </summary>
		/// <returns></returns>
		size_t getFloatDataSize() const;

		///@}


		///@{
		///@name Other Properties

		/// <summary>
		/// Row stride of the image.  For images with 8 or 16 bits per sample, this is in samples; for 32-bit images, it's in bytes.
		/// </summary>
		/// 32-bit images are stored in RGBE format four bytes per sample.  You can call Image::convertToFP() to get RGB data with 32-bit floating point data per channel.
		size_t getRowStride() const noexcept;
		
		/// <summary>
		/// Row stride of the image in bytes
		/// </summary>
		size_t getRowStrideBytes() const noexcept;
		
		/// <summary>
		/// Size of the image data in bytes
		/// </summary>
		size_t getImageStrideBytes() const noexcept;

		/// <summary>
		/// Resolution in the X dimension of the image in pixels per inch
		/// </summary>
		int getXres() const noexcept;

		/// <summary>
		/// Resolution in the Y dimension of the image in pixels per inch
		/// </summary>
		int getYres() const noexcept;

		/// <summary>
		/// Orientation to use when displaying the image
		/// </summary>
		EImageOrientation getOrientation() const noexcept;

		/// <summary>
		/// True if the image has EXIF data associated with it
		/// </summary>
		bool hasEXIFData() const noexcept;

		/// <summary>
		/// Set the color profile that will be used when loading a RAW image.
		/// </summary>
		/// This function is ignored if the image is not RAW.  It must be called before loading the image.
		/// Normally you would use the PhotomatixEngine::setRAWColorProfile() method to apply the same color profile to all RAW images.
		/// You need this function only if you want to set the colour profile for each image differently.
		/// <param name="profile">Color profile to use</param>
		void setRAWColorProfile(EColorProfile profile) noexcept;

		/// <summary>
		/// The color profile that will be used when loading a RAW image.
		/// </summary>
		EColorProfile getRAWColorProfile() const noexcept;

		/// <summary>
		/// Set the color temperature that will be used whan loading a RAW image.  It must be called before loading the image.
		/// </summary>
		/// This function is ignored if the image is not RAW.  Normally you would use the
		/// PhotomatixEngine::setRAWColorTemperature() method to apply the same color temperature to all RAW images.
		/// You need this function only if you want to set the colour temperature for each image differently.
		/// <param name="colorTemp">Color temperature to use</param>
		void setRAWColorTemperature(int colorTemp) noexcept;

		/// <summary>
		/// The color temperature to use when loading a RAW image.
		/// </summary>
		/// This returns 0 if the color temperature to use is not set, in which "as shot" will be used, or it will be set based on the temperature choices
		/// made when calling PhotomatixEngine::setRAWColorTemperature().
		int getRAWColorTemperature() const noexcept;

		/// <summary>
		/// Turn on/off automatic brightness adjustment when loading RAW images.  It must be called before loading the image.
		/// </summary>
		/// This function is ignored if the image is not RAW.
		/// Use the PhotomatixEngine::setRAWAutoBrightness() method to apply this setting to all RAW images unless you want only some images to use auto brightness.
		/// <param name="autoBrightness">When true, the image brightness is adjusted to preserve highlights when loading (default).
		/// When false, no adjustments are made, which can be useful when processing video frames which must have consistent brightness.</param>
		void setRAWAutoBrightness(bool autoBrightness) noexcept;

		/// <summary>
		/// Whether automatic brightness adjustment will be used when loading RAW images.
		/// </summary>
		bool getRAWAutoBrightness() const noexcept;


		/// <summary>
		/// Turn on/off highlight recovery when loading RAW images.
		/// </summary>
		/// If highlight recovery is turned on, images may be significantly darker.  It's recommended to turn this on
		/// only if processing a single image.
		/// Use the PhotomatixEngine::setRAWHighlightRecovery() method to apply the same setting to all RAW images unless you want only some images to have highlight recovery.
		void setRAWHighlightRecovery(bool highlightRecovery) noexcept;

		/// <summary>
		/// Whether highlight recovery will be used when loading this image.
		/// </summary>
		bool getRAWHighlightRecovery() const noexcept;

#ifdef DENOISE_INCLUDED
		/// <summary>
		/// Turn on/off noise reduction for this image.
		/// </summary>
		/// Normally you would use PhotomatixEngine::setNoiseReductionType() to automatically turn on/off noise reduction on loaded images based on exposure type.
		/// This function is useful if you want to control exactly which images have noise reduction, and the categories available for PhotomatixEngine::setNoiseReductionType() aren't suitable.
		void setNoiseReduction(bool noiseReduction) noexcept;

		/// <summary>
		/// True if noise reduction is to be done on this image when loaded.
		/// </summary>
		bool getNoiseReduction() const noexcept;

		/// <summary>
		/// True if noise reduction has been manually turned on/off for this image, rather than it being set from a call to PhotomatixEngine::setNoiseReductionType().
		/// </summary>
		/// <returns></returns>
		bool noiseReductionOverridden() const noexcept;
#endif

#ifdef LENS_CORRECT_INCLUDED
		class LensList;

		/// <summary>
		/// Get the collection of lenses that are applicable to this image for lens correction, based on its metadata.
		/// </summary>
		/// The lenses listed are either those which exactly match the lens in the image metadata, or if there are no exact matches, all those which are compatible with the
		/// camera, aperture and focal length read from the image metadata.  You can query LensList::exactMatch() in the returned object to find whether this is a list of exact matches or not.
		LensList getMatchingLenses() const;

		/// <summary>
		/// Set the lens to use for lens correction of the image.
		/// </summary>
		/// The lens description provided must be one returned by getMatchingLensWithIndex() const.
		/// Throws PhotomatixException if the lens description is not valid.
		/// Note that if this is not set, but PhotomatixEngine::setAutoLensCorrection() is turned on, lens correction will still be done on loading, if there is a single exactly matching lens.
		void setLensForCorrection(const char *lensForCorrection) noexcept;

		/// <summary>
		/// Returns the description of the lens selected for correction, or nullptr if none has been selected.
		/// </summary>
		const char* getLensForCorrection() const noexcept;

		///@}


		/// <summary>
		/// Iterator for collection of lenses that could be used for lens correction for an image.  Obtained by calling begin() and end() on the collection returned from Image::getMatchingLenses().
		/// </summary>
		class EXPORT LensIterator
		{
		public:
			LensIterator& operator++() noexcept;
			LensIterator operator++(int) noexcept;
			LensIterator& operator--() noexcept;
			LensIterator operator--(int) noexcept;
			const char* operator*() const noexcept;
			bool operator==(const LensIterator& rhs) const noexcept;
			bool operator!=(const LensIterator& rhs) const noexcept;

			typedef std::bidirectional_iterator_tag iterator_category;
			typedef const char* value_type;
			typedef std::ptrdiff_t difference_type;
			typedef const char** pointer;
			typedef const char *& reference;
		private:
			friend class LensList;
			LensIterator(const void* pos) noexcept : idx(pos) { }
			const void* idx;
		};


		/// <summary>
		/// Collection of lenses that could be used for lens correction for an image, obtained by calling Image::getMatchingLenses()
		/// </summary>
		class EXPORT LensList
		{
		public:
			/// <summary>
			/// Returns start iterator
			/// </summary>
			LensIterator begin() const noexcept;
			/// <summary>
			/// Returns end iterator
			/// </summary>
			LensIterator end() const noexcept;
			/// <summary>
			/// Number of lenses in collection
			/// </summary>
			size_t size() const noexcept;
			/// <summary>
			/// Lens with the specified index.
			/// </summary>
			const char* operator[](size_t index) const;
			/// <summary>
			/// If true, the lens(es) returned are exact matches based on the image metadata;
			/// otherwise there were no exact matches, and the lenses are those which are compatible with the camera, aperture and focal length of the image.
			/// </summary>
			/// <returns></returns>
			bool exactMatch() const noexcept;
			using iterator = LensIterator;
			using const_iterator = LensIterator;
		private:
			friend class Image;
			LensList(const Image& img) noexcept : imgRef(img) { }
			const Image& imgRef;
		};
#endif

		/// \cond PRIVATE
		~Image() noexcept;
		Image(const Image&);
		Image(Image&&) noexcept;
		Image& operator=(const Image&);
		Image& operator=(Image&&) noexcept;
		/// \endcond
	
	private:
		friend class PhotomatixEngine;
		Image();
		Image(const PhotomatixEngine *engine, int width, int height, int bps, int spp, size_t rowStride, const unsigned char* data);
		static Image createFromData(const PhotomatixEngine *engine, int width, int height, int bps, int spp, size_t rowStride, const unsigned char* data, bool isRGBE = false);
		static Image createFromFile(const PhotomatixEngine *engine, const char* fileName, bool createFakeExposures);
		bool hasData() const noexcept;
		void getSaveInfos(int saveInfos[]) const noexcept;
		void createLensList() const;
		int* getExifBuf() const noexcept;
		void setExifBuf(const int* exifBuf) noexcept;
		int getFileType() const noexcept;
		void setOrientation(EImageOrientation orientation) noexcept;
		void setMetadataSource(const char* src);
		const char* getMetadataSource() noexcept;
		unsigned char* getColorProfileData() const noexcept;
		size_t getColorProfileLen() const noexcept;
		void setColorProfileData(size_t size, const unsigned char *data);
		float getRawColorScalingNoHL() const noexcept;
		const unsigned short* getSrgbCurve() const noexcept;
		void moveData(unsigned char* dest);
		void setThreads(int threads) noexcept;
		void changeDimensions(int width, int height, size_t rowStride, unsigned char* engineData) noexcept;
		void lensCorrect(const char* lensDescription);
		
		struct ImageInstanceData;
		ImageInstanceData* instanceData = nullptr;
	};
}

#endif
