//
//  ContentView.swift
//  PhotomatixLibTest
//
//  Created by Matthew Donaldson on 19/11/2024.
//

import SwiftUI
import PhotomatixLib

struct ContentView: View {
    @State private var showPicker = false
    @State private var selectedImages: [UIImage] = []
    @State private var selectedFilePaths: [String] = []
    
    @State private var processedImage: UIImage? // State to hold the processed image
    @State private var reprocessedImage: UIImage? // State to hold the reprocessed image
    
    var body: some View {
        VStack {
            Image(systemName: "globe")
                .imageScale(.large)
                .foregroundStyle(.tint)
            Text("Hello, world!")
            Button("Pick the photo(s)") {
                showPicker = true
            }.padding()
            
            // Display selected images
            ScrollView(.horizontal) {
                HStack {
                    ForEach(selectedImages, id: \.self) { image in
                        Image(uiImage: image)
                            .resizable()
                            .scaledToFit()
                            .frame(width: 100, height: 100)
                            .clipped()
                    }
                }
            }
            .padding()
            
            // Display file paths
            ScrollView(.vertical) {
                VStack(alignment: .leading) {
                    ForEach(selectedFilePaths, id: \.self) { path in
                        Text(path)
                            .font(.footnote)
                            .foregroundColor(.gray)
                    }
                }
            }
            .padding()
            
            if !selectedFilePaths.isEmpty {
                // Display Process File button if there are files selected
                Button("Process File") {
                    // All the processing with PhotomatixLib is in here
                    if let img = getProcessedImage(selectedFilePaths) {
                        processedImage = img
                    }
                }
                
                if let uiImage = processedImage {
                    Image(uiImage: uiImage) // Display the processed UIImage
                        .resizable()
                        .scaledToFit()
                        .frame(maxWidth: 300, maxHeight: 300) // Adjust the size as needed
                        .padding()
                    Button("Reprocess") {
                        if let newImage = reprocessImage(image: uiImage) {
                            reprocessedImage = newImage
                        }
                }
                        
                if let uiImage2 = reprocessedImage {
                    Image(uiImage: uiImage2)
                        .resizable()
                        .scaledToFit()
                        .frame(maxWidth: 300, maxHeight:300)
                        .padding()
                    }
                }
            }
        }
        .sheet(isPresented: $showPicker) {
            PhotoPicker(selectedImages: $selectedImages, selectedFilePaths: $selectedFilePaths)
        }
    }
    
  
    func getProcessedImage(_ filePaths: [String]) -> UIImage? {
        // Create a C++ shared pointer to a PhotomatixEngine, which is responsible for storing source images, then merging them.
        // Creating a shared pointer is better than constructing a PhotomatixEngine explicitly, as Swift sometimes copies the object itself
        // when calling methods, and then doesn't always handle the lifetime of the copied object appropriately.
        // The Swift/C++ interop manages the memory of the PhotomatixEngine object, deleting it when all shared_ptr references go out of scope.
        // The shared_ptr object can be treated similarly to a normal Swift pointer, but explicitly needs .pointee to reference the obect itself.
        let engine = photomatix.PhotomatixEngine.create()
        
        // Turn on alignment of images, using the normal method handling both rotation and perspective variations.
        engine.pointee.setAlignType(photomatix.EAlignmentType.RotationPerspective)

        // You can set a callback for progress logs if you want to do things like a progress bar or something.
        // The callback is passed a message, a progress proportion from 0.0 to 1.0 if the operation is long, and optionally a user data object
        // you specify when setting the callback, which could be a reference to a class object for example.
        engine.pointee.setProgressLogger({ (message, progress, userData) in
            if let msg = message {
                let messageString = String(cString: msg)
                print("> Progress: \(messageString), \(progress * 100)%")
        } }, nil, false)
        
        let fileManager = FileManager.default
        do {
            for filePath in filePaths {
                if fileManager.fileExists(atPath: filePath) {
                    // Add the image to the engine if it exists.
                    // This is a helper function that converts C++ exceptions to Swift exceptions, as Swift can't handle C++ exceptions.
                    // This calls engine.addImageFromFile(), which should not be called directly from Swift as if it throws an
                    // exception it will crash the app, since Swift can't catch it.
                    // This returns a pointer to the created Image object, but we don't generally need it.
                    _ = try PMImageUtils.addImage(toEngine: engine, fromFile: filePath)

                    // There is also a method in PhotomatixEngine which handles exceptions by returning a struct with the image, and an exception
                    // message.  If the image property is null then there should be an exception string in the message property.  This is an
                    // alternate way to add images, not generating exceptions at all.
                    //                let imageResult =
                    //                engine.pointee.addImageFromFileWithEx(filePath, false)
                    //                if imageResult.image == nil {
                    //                    print("Error: \(imageResult.message)")
                    //                    return nil
                }
            }

            // There are a number of built in presets, with the xmp extension.  Other presets can be created using the Photomatix Pro application,
            // and saved as xmp files in a place accessible to the app.
            guard let filePath = Bundle.main.path(forResource: "Natural", ofType: "xmp") else {
                print("Couldn't find preset.")
                return nil
            }

            let count = engine.pointee.getImages().size()
            print("Number of images: \(count)")

            // Turn on other options such as noise reduction
            // engine.pointee.setNoiseReductionType(photomatix.ENoiseReductionType.NormalAndUnderExposed)

            // Create the HDRPreset object from the xmp file.
            let preset = photomatix.HDRPreset.createFromFile(filePath)

            print("Processing \(count) image(s) with a preset with method \(preset.getMethod())")
            // This is another helper function that converts C++ exceptions to Swift exceptions.
            // It calls engine.process() under the hood, which also shouldn't be called directly from Swift as it can throw C++ exceptions.
            // If successful, it returns a C++ shared_ptr to a photomatix.Image object containing the merged and processed image,
            // otherwise it throws a Swift exception.
            let newImage = try PMImageUtils.process(withEngine: engine, preset: preset, isPanorama: false).image
            print("---Finished process---")
        
            print("---About to call uiImage---")
            // This helper function converts from the photomatix.Image object to a UIImage.
            if let img = PMImageUtils.uiImage(fromImagePtr: newImage)
            {
                print("Successfully created UIImage")
                return img
            }
            return nil
        } catch let error as PMOutOfMemoryException {
            print("Out of Memory: \(error.localizedDescription)")
            return nil

        } catch let error as PMInvalidImageException {
            print("Image Exception: \(error.localizedDescription)")
            return nil

        } catch let error as PMPhotomatixException {
            print("Photomatix Exception: \(error.localizedDescription)")
            return nil

        } catch {
            print("Unknown Error: \(error.localizedDescription)")
            return nil
        }
    }
    
    func reprocessImage(image : UIImage) -> UIImage? {
        let engine = photomatix.PhotomatixEngine.create()
        do {
            // Add the image to the engine.
            // This is a helper function that adds the data in a UIImage, and converts C++ exceptions to Swift exceptions, as Swift can't handle C++ exceptions.
            // This returns a pointer to the created Image object, but we don't generally need it.
            _ = try PMImageUtils.addImage (toEngine: engine, from: image)

            // Load the Detailed preset
            guard let filePath = Bundle.main.path(forResource: "Detailed", ofType: "xmp") else {
                print("Couldn't find preset.")
                return nil
            }

            // Create the HDRPreset object from the xmp file.
            let preset = photomatix.HDRPreset.createFromFile(filePath)

            print("Processing image with a preset with method \(preset.getMethod())")

            // Process the image with this preset
            let newImage = try PMImageUtils.process(withEngine: engine, preset: preset, isPanorama: false).image
            print("---Finished process---")
        
            print("---About to call uiImage---")
            // This helper function converts from the photomatix.Image object to a UIImage.
            if let img = PMImageUtils.uiImage(fromImagePtr: newImage)
            {
                print("Successfully created UIImage")
                return img
            }
            return nil
        } catch let error as PMOutOfMemoryException {
            print("Out of Memory: \(error.localizedDescription)")
            return nil

        } catch let error as PMInvalidImageException {
            print("Image Exception: \(error.localizedDescription)")
            return nil

        } catch let error as PMPhotomatixException {
            print("Photomatix Exception: \(error.localizedDescription)")
            return nil

        } catch {
            print("Unknown Error: \(error.localizedDescription)")
            return nil
        }
    }
}

#Preview {
    ContentView()
}
