//
//  ImagingUtils.swift
//  PhotomatixLibTest
//
//  Created by Matthew Donaldson on 16/12/2024.
//

import Foundation
import PhotomatixLib

// These functions aren't used in the example code, they are just showing how to convert from a photomatix.Image to a UIImage in Swift.
// There is also library code in the ImageUtils Objective C++ wrapper which will do the same thing.
func mapOrientation(_ orientation: photomatix.EImageOrientation) -> UIImage.Orientation {
    switch orientation {
    case .normal:
        return .up
    case .flippedHorizontally:
        return .upMirrored
    case .upsideDown:
        return .down
    case .flippedVertically:
        return .downMirrored
    case .flippedHorizRot270Deg:
        return .leftMirrored
    case .rotated90Clockwise:
        return .right
    case .flippedVertRot90Deg:
        return .rightMirrored
    case .rotated270Clockwise:
        return .left
    @unknown default:
        return .up // Fallback for unknown values
    }
}

func createUIImage(from image: photomatix.Image) -> UIImage? {
    let width = Int(image.getWidth())
    let height = Int(image.getHeight())
    let bitsPerComponent = Int(image.getBps()) // Bits per sample
    let componentsPerPixel = Int(image.getSpp()) // Samples per pixel (e.g., 4 for RGBA)
    let bytesPerRow = Int(image.getRowStride()) // Stride for one row of pixels
    
    
    guard let dataPointer = image.__getDataUnsafe() else {
        print("Failed to retrieve image data")
        return nil
    }
    
    // Copy raw data into a Swift Data object
    let dataSize = bytesPerRow * height
    let imageData = Data(bytes: dataPointer, count: dataSize)
    
    // Create a CGDataProvider for the raw data
    guard let provider = CGDataProvider(data: imageData as CFData) else {
        print("Failed to create CGDataProvider")
        return nil
    }
    
    // Assume the image format is RGBA with 8 bits per component
    let bitmapInfo = CGBitmapInfo(rawValue: CGImageAlphaInfo.none.rawValue)
    
    // Create a CGImage from the raw data
    guard let cgImage = CGImage(
        width: width,
        height: height,
        bitsPerComponent: bitsPerComponent,
        bitsPerPixel: bitsPerComponent * componentsPerPixel,
        bytesPerRow: bytesPerRow,
        space: CGColorSpaceCreateDeviceRGB(),
        bitmapInfo: bitmapInfo,
        provider: provider,
        decode: nil,
        shouldInterpolate: true,
        intent: .defaultIntent
    ) else {
        print("Failed to create CGImage")
        return nil
    }
    
    let uiImageOrientation = mapOrientation(image.getOrientation())
    return UIImage(cgImage: cgImage, scale: 1.0, orientation: uiImageOrientation)
}
