import SwiftUI
import PhotosUI

struct PhotoPicker: UIViewControllerRepresentable {
    @Binding var selectedImages: [UIImage]
    @Binding var selectedFilePaths: [String]

    func makeUIViewController(context: Context) -> PHPickerViewController {
        var config = PHPickerConfiguration()
        config.filter = .images
        config.selectionLimit = 0 // Allow multiple selection

        let picker = PHPickerViewController(configuration: config)
        picker.delegate = context.coordinator
        return picker
    }

    func updateUIViewController(_ uiViewController: PHPickerViewController, context: Context) {}

    func makeCoordinator() -> Coordinator {
        Coordinator(self)
    }

    class Coordinator: NSObject, PHPickerViewControllerDelegate {
        let parent: PhotoPicker

        init(_ parent: PhotoPicker) {
            self.parent = parent
        }

        func picker(_ picker: PHPickerViewController, didFinishPicking results: [PHPickerResult]) {
            parent.selectedImages.removeAll()     // Clear previous results
            parent.selectedFilePaths.removeAll() // Clear previous file paths

            let dispatchGroup = DispatchGroup()

            for result in results {
                if result.itemProvider.hasItemConformingToTypeIdentifier(UTType.image.identifier) {
                    dispatchGroup.enter()

                    // Extract UIImage
                    result.itemProvider.loadObject(ofClass: UIImage.self) { image, error in
                        if let image = image as? UIImage {
                            DispatchQueue.main.async {
                                self.parent.selectedImages.append(image)
                            }
                        } else {
                            print("Error loading image: \(String(describing: error))")
                        }
                    }

                    // Extract File URL
                    result.itemProvider.loadFileRepresentation(forTypeIdentifier: UTType.image.identifier) { url, error in
                        if let url = url {
                            do {
                                // Create unique file path in temp dir
                                let fileManager = FileManager.default
                                let tempURL = fileManager.temporaryDirectory.appendingPathComponent(url.lastPathComponent)

                                // Copy file to temp location
                                if fileManager.fileExists(atPath: tempURL.path) {
                                    try fileManager.removeItem(at: tempURL)
                                }
                                try fileManager.copyItem(at: url, to: tempURL)
                                DispatchQueue.main.async {
                                    self.parent.selectedFilePaths.append(tempURL.path)
                                }
                            } catch {
                                print("Error copying file: \(error.localizedDescription)")
                            }
                        }
                        dispatchGroup.leave()
                    }
                }
            }

            // Dismiss picker after processing
            dispatchGroup.notify(queue: .main) {
                picker.dismiss(animated: true)
            }
        }
    }
}
