#ifndef _PHOTOMATIX_HDRPRESET_H
#define _PHOTOMATIX_HDRPRESET_H

//#include "../Photomatix32.h"

#if defined(_WIN32) && !defined(SWIG)
#define EXPORT __declspec(dllexport)
#else
#define EXPORT
#endif

/// \file HDRPreset.h
/// \brief Contains the definition for the HDRPreset class, which stores settings to be used for HDR processing.
/// 

namespace photomatix
{

	const int XMP_LEN = 60;

	/// <summary>
	/// The different methods, or image processing algorithms available
	/// </summary>
	enum class EMethodType
	{
		None = 0,
		/// <summary>
		/// The Details Enhancer tone mapping algorithm
		/// </summary>
		DetailsEnhancer = 1,
		/// <summary>
		/// The Tone Compressor tone mapping algorithm
		/// </summary>
		ToneCompressor = 2,
		/// <summary>
		/// The Contrast Optimizer tone mapping algorithm
		/// </summary>
		ContrastOptimizer = 3,
		/// <summary>
		/// The Tone Balancer tone mapping algorithm
		/// </summary>
		ToneBalancer = 4,
		/// <summary>
		/// The Auto fusion algorithm
		/// </summary>
		Auto = 12,
		/// <summary>
		/// The Average fusion algorithm, which just averages the images together
		/// </summary>
		Average = 14,
		/// <summary>
		/// Two Images algorithm, not used in this library
		/// </summary>
		TwoImages = 13,
		/// <summary>
		/// The Natural fusion algorithm
		/// </summary>
		Adjust = 10,
		/// <summary>
		/// The Intensive fusion algorithm
		/// </summary>
		Intensive = 11,
		/// <summary>
		///  The Realistic real-estate specialist fusion algorithm
		/// </summary>
		Realistic = 15
	};

	/// <summary>
	/// Stores settings for HDR processing, either read from a file, or constructed from a method and settings for that method.  Presets are
	/// used for specifying the settings for HDR processing.
	/// </summary>
	class EXPORT HDRPreset
	{
	public:
		/// <summary>
		/// Load a preset from a file, or throws a PhotomatixException if it can't be loaded.
		/// </summary>
		/// <param name="xmpFile"></param>
		/// <returns>Loaded preset</returns>
		static HDRPreset createFromFile(const char* xmpFile);

		/// <summary>
		/// Create a preset from a method type and parameters for that method
		/// </summary>
		/// <param name="method">Method, specifying the processing algorithm to use</param>
		/// <param name="params">An array of parameters for the method, consisting of 60 float values.  Not all values are used for all methods.</param>
		/// <returns>Created preset</returns>
		static HDRPreset createFromSettings(EMethodType method, const float params[XMP_LEN]);

		/// <summary>
		/// Create a preset from a string containing the XML preset data.
		/// </summary>
		/// <param name="xmlString">Null-terminated string with the preset data</param>
		/// <param name="name">Optional name of the preset, for display purposes.</param>
		/// <returns>Created preset</returns>
		static HDRPreset createFromXMLString(const char* xmlString, const char *name = "");

		/// <summary>
		/// Get the method for this preset.
		/// </summary>
		EMethodType getMethod() const noexcept;

		/// <summary>
		/// Does this preset include blending the processed result with the original unprocessed image(s)?
		/// </summary>
		bool hasBlendWithOriginal() const noexcept;

		/// <summary>
		/// True if this is a fusion preset, which directly merges images together without first creating a 32-bit HDR image.
		/// </summary>
		bool isFusion() const noexcept;

		/// <summary>
		/// True if this is a tone mapping preset, which first creates a 32-bit HDR image, before creating a standard LDR image from it.
		/// </summary>
		bool isToneMapping() const noexcept;

		/// <summary>
		/// The parameters for this preset, as an array of 60 floats.
		/// </summary>
		float* getParams() const noexcept;
		
		/// <summary>
		/// Printable name for this preset, if it was loaded from a file.  The empty string otherwise.
		/// </summary>
		const char* getName() const noexcept;

		/// <summary>
		/// Returns true if this preset has the same method and parameters as other.
		/// </summary>
		bool hasSameSettingsAs(const HDRPreset& other) const noexcept;

		/// \cond PRIVATE
		~HDRPreset() noexcept;
		HDRPreset(const HDRPreset& other);
		HDRPreset(HDRPreset&& other) noexcept;
		HDRPreset& operator=(const HDRPreset& rhs);
		HDRPreset& operator=(HDRPreset&& rhs) noexcept;
		/// \endcond

	private:
		HDRPreset();
		struct HDRPresetInstanceData;
		HDRPresetInstanceData *instanceData = nullptr;
	};
}

#endif
