#ifndef PHOTOMATIX_PHOTOMATIXENGINE_H
#define PHOTOMATIX_PHOTOMATIXENGINE_H

#ifdef __INTBUILD
#include "../Features.h"
#else
#include "Features.h"
#endif

#include <vector>
#include "Image.h"
#include "HDRPreset.h"

#ifndef OLD_CPLUSPLUS
#ifdef _WIN32
#define OLD_CPLUSPLUS (_MSC_VER < 1900)
#else
#define OLD_CPLUSPLUS (__cplusplus <= 199711L)
#endif
#if OLD_CPLUSPLUS
#define noexcept throw()
#endif
#endif


#if defined(_WIN32) && !defined(SWIG)
#define EXPORT __declspec(dllexport)
#else
#define EXPORT
#endif

/// \file PhotomatixEngine.h
/// \brief The header file for the PhotomatixEngine class, which is responsible for loading and merging multiple images to HDR.
/// 

namespace photomatix
{
	/// <summary>
	/// Enumeration describing which source photos (if any) to apply noise reduction to
	/// </summary>
	enum class ENoiseReductionType
	{
		None = -1,		///< Do no noise reduction
		UnderExposed = 0, ///< Do noise reduction only on underexposed photos
		NormalAndUnderExposed = 1, ///< Do noise reduction on the normal (median) exposure and underexposed photos
		All = 2 ///< Do noise reduction on all photos
	};

	/// <summary>
	/// Enumeration describing what type of deghosting to perform
	/// </summary>
	enum class EDeghostType
	{
		/// <summary>
		/// Do no deghosting
		/// </summary>
		None = 0,
		/// <summary>
		/// Automatically detect ghosted areas and remove ghosts
		/// </summary>
		Automatic = 1,
		/// <summary>
		/// INTERNAL USE
		/// </summary>
		SemiManual = 2  // Remove ghosts in areas supplied as a bitmask file
	};

	/// <summary>
	/// Enumeration describing the type of image alignment to perform with bracketed sets
	/// </summary>
	enum class EAlignmentType
	{
		/// <summary>
		/// No image alignment
		/// </summary>
		None = -1,
		/// <summary>
		/// Align using only translation movement of images
		/// </summary>
		Translation = 0,
		/// <summary>
		/// Align using both translation, rotation and perspective correction of images.  This is the recommended setting.
		/// </summary>
		RotationPerspective = 1,
		/// <summary>
		/// Align using translation and rotation
		/// </summary>
		Rotation = 2

	};

	/// <summary>
	/// Enumeration describing the RAW images which color temperature settings apply to.
	/// </summary>
	enum class EColorTempSelection
	{
		/// <summary>
		/// Apply color temperature to only underexposed photos
		/// </summary>
		UnderExposed,
		/// <summary>
		/// Apply color temperature to only the normal (median) exposure
		/// </summary>
		Normal,
		/// <summary>
		/// Apply color temperature to only overexposed photos
		/// </summary>
		OverExposed,
		/// <summary>
		/// Apply color temperature to all photos
		/// </summary>
		All
	};

	/// <summary>
	/// Enumeration describing where to get the tonal curve from when loading RAW images
	/// </summary>
	enum class ETonalCurveSource
	{
		/// <summary>
		/// Use ICC profile of source images
		/// </summary>
		ICCProfile,
		/// <summary>
		/// Attempt to recover curve from the image itself
		/// </summary>
		Image
	};

	/// <summary>
	/// The main class that loads images, merges them, and generates result images after processing with a preset.
	/// </summary>
	/// This is the class you start with when merging images and processing them, and when enhancing single images. After constructing a PhotomatixEngine object, you
	/// add your source file(s) to it, set the options you want for how to load and merge the source files, such as alignment and noise reduction, then get
	/// the 32-bit HDR merged image, or merge and process the source images with settings in an HDRPreset object.
	/// See @ref MergeAndProcess.cpp for a simple example of using this class.
	class EXPORT PhotomatixEngine
	{
	public:

		///@{
		/// @name Constructor

		/// <summary>
		/// Construct a PhotomatixEngine object.
		/// </summary>
		PhotomatixEngine();

		///@}

		///@{
		///@name Adding Source Files

		/// <summary>
		/// Add an image from a data buffer.  Copies the buffer into the object.
		/// </summary>
		/// The data supplied must be with interleaved channels - for example, if there are three samples per pixel (or channels), then the pixel data must be stored like this: RGBRGBRGB...
		/// If there are more than three three samples per pixel, for example, if there is an alpha channel, then the red, green and blue data must be the first three samples in each pixel.
		/// If bps is 8, then the samples are expected to be unsigned byte values; for bps of 16, unsigned short values are expected; and for bps of 32, 32-bit float values must be supplied for each sample.
		/// The buffer **must** be at least the size of the row stride multiplied by the height.
		/// <param name="width">Width of the image in pixels</param>
		/// <param name="height">Height of the image in pixels</param>
		/// <param name="bps">Bits per sample, must be 8, 16 or 32</param>
		/// <param name="spp">Samples per pixel, or channels in the image.  Must be at least 3.</param>
		/// <param name="rowStride">Row stride in samples (not bytes).  If this is greater than the minimum size required for the pixels in a row, then extra padding is added at the end of each row.  Pass in zero to use the minimum row stride.</param>
		/// <param name="data">The data for the image, with samples for each channel interleaved</param>
		/// <returns>Pointer to a new Image object, owned by the PhotomatixEngine object.  It will be destructed along with its PhotomatixEngine object.</returns>
		Image *addImageFromData(int width, int height, int bps, int spp, size_t rowStride, const unsigned char* data);

		/// <summary>
		/// Add an image by loading from a file.  Throws an InvalidImage exception if it can't be read.
		/// </summary>
		/// Note that this reads only the metadata of the image.  The image data itself can be loaded explicitly by calling Image::load() if any preprocessing
		/// needs to be done to it, such as lens correction or contrast enhancement, or you need to access the image data by calling Image::getData().
		/// Otherwise it will be loaded automatically when processing is done.
		/// <param name="fileName">Path to file</param>
		/// /// <param name="createFakeExposures">Create fake exposures when loading a RAW image.  When true, multiple photos at different exposures are created from the loaded image, so that
		/// Fusion presets can be used with a single image.  The image must be RAW for this to be done.  Only a single image may be added to the PhotomatixEngine if this parameter is true.</param>
		/// <returns>Pointer to a new Image object, owned by the PhotomatixEngine object.  It will be destructed along with its PhotomatixEngine object.</returns>
		Image *addImageFromFile(const char* fileName, bool createFakeExposures = false);

		///@}

#ifdef ALIGN_INCLUDED
		///@{
		///@name Image Alignment Options

		/// <summary>
		/// Turn on alignment of images during HDR generation if there are multiple images, and specify which type should be used
		/// </summary>
		/// <param name="alignType">Which alignment type should be used</param>
		void setAlignType(EAlignmentType alignType) noexcept;

		/// <summary>
		/// The type of image alignment to be done during HDR generation if there are multiple images, or EAlignmentType_None if no alignment is to be done.
		/// </summary>
		EAlignmentType getAlignType() const noexcept;


		/// <summary>
		/// Set the maximum shift of images with respect to one another when trying to align them, as a percentage of image size.  Defaults to 12%.
		/// </summary>
		void setMaxAlignmentShift(int maxAlignmentShift) noexcept;

		/// <summary>
		/// The maximum shift of images with respect to one another when trying to align them, as a percentage of image size
		/// </summary>
		/// <returns></returns>
		int getMaxAlignmentShift() const noexcept;

		///@}
#endif


#ifdef DEGHOST_INCLUDED
		///@{
		///@name Ghost Removal Options


		/// <summary>
		/// Set the type of deghosting to be done during merging of images.  Defaults to EDeghostingType::None: no deghosting.
		/// </summary>
		void setDeghostingType(EDeghostType deghostingType) noexcept;

		/// <summary>
		/// Type of deghosting to be done when merging images, or EDeghostingType::None if no deghosting is to be done.
		/// </summary>
		EDeghostType getDeghostingType() const noexcept;

		/// <summary>
		/// Set strength for automatic deghosting.  Valid values are between 0 and 100, default is 50.  Higher values remove more ghosting, but can create image artifacts.
		/// </summary>
		/// <param name="deghostingStrength">Deghosting strength.  There are two predefined constants that can also be used:
		///   PhotomatixEngine::DeghostStrengthNormal and PhotomatixEngine::DeghostStrengthHigh.</param>
		/// Regardless of this setting, ghost removal will not be done unless it is turned on by calling setDeghostingType().
		void setDeghostingStrength(int deghostingStrength) noexcept;

		/// <summary>
		/// Returns the strength for automatic deghosting.
		/// </summary>
		int getDeghostingStrength() const noexcept;

		/// <summary>
		/// Set EV offset of base exposure in the supplied bracketed set to use for automatic deghosting.
		/// </summary>
		/// Positive values select more overexposed images,
		/// negative values select more underexposed images, and zero (the default) uses the median exposure.  The image with the nearest offset to the one supplied will be used.
		void setDeghostingBaseExposure(float deghostingBaseExposure) noexcept;

		/// <summary>
		/// Get EV offset of base exposure to use for automatic deghosting.
		/// </summary>
		/// <returns></returns>
		float getDeghostingBaseExposure() const noexcept;

		///@}

		/// <summary>
		/// Convenience constant to pass to setDeghostingStrength for normal strength deghosting.  This is the default, so setting this is not normally necessary.
		/// </summary>
		static const int DeghostStrengthNormal = 50;

		/// <summary>
		/// Convenience constant to pass to setDeghostingStrength for high strength deghosting.  This may resolve more ghosting artifacts, but also runs the risk of introducing other artifacts.
		/// </summary>
		static const int DeghostStrengthHigh = 90;
#endif


#ifdef DENOISE_INCLUDED
		///@{
		///@name Noise Reduction Options

		/// <summary>
		/// Set the type of noise reduction to be done on source images, if any.
		/// See ENoiseReductionType for valid values.
		/// </summary>
		void setNoiseReductionType(ENoiseReductionType noiseReductionType) noexcept;

		/// <summary>
		/// What type of noise reduction is to be done on source images, if any.
		/// </summary>
		ENoiseReductionType getNoiseReductionType() const noexcept;

		/// <summary>
		/// Enable/disable noise reduction for the generated 32-bit HDR image created from merged images.
		/// </summary>
		void setHDRNoiseReductionEnabled(bool noiseReductionEnabled) noexcept;

		/// <summary>
		/// Is noise reduction for the generated 32-bit HDR image created from merged images turned on or off?
		/// </summary>
		bool getHDRNoiseReductionEnabled() const noexcept;

		// Noise reduction strength (50 to 150, default 100)

		/// <summary>
		/// Strength of noise reduction (valid values are 50 to 150, default 100).
		/// </summary>
		void setNoiseReductionStrength(int noiseReductionStrength) noexcept;

		/// <summary>
		/// Strength of noise reduction
		/// </summary>
		int getNoiseReductionStrength() const noexcept;

		///@}
#endif

		///@{
		///@name Other Common Merging and Loading Options

#ifdef LENS_CORRECT_INCLUDED
		/// <summary>
		/// Turn on/off automatic lens correction.
		/// </summary>
		/// When turned on, if there is a single exact match for the lens needed to do lens correction for an image,
		/// then lens correction will be performed.  However if a call to Image::setLensForCorrection() has explicitly set a lens description to use for a particular image,
		/// that description will take precedence.
		void setAutoLensCorrection(bool autoLensCorrection) noexcept;

		/// <summary>
		/// Whether automatic lens correction is enabled.
		/// </summary>
		/// When turned on, if there is a single exact match for the lens needed to do lens correction for an image,
		/// then lens correction will be performed, unless a lens had already been set on the image using Image::GetLensForCorrection().
		bool getAutoLensCorrection() const noexcept;
#endif

		///@}

		using LogFunc = void(const char* operation, float progress, void* userData);
		///@{
		///@name Getting Progress Notifications

		/// <summary>
		/// Set a function to be called whenever a progress log message is written.
		/// </summary>
		/// The name of the current operation, and the proportion through from 0.0 to 1.0 (if applicable) are passed to the log function,
		/// as well as the user data that was supplied, if any.
		/// If the operation doesn't provide any progress information, the proportion through is passed as 0.
		/// <param name="logger">Log function</param>
		void setProgressLogger(LogFunc* logger, void* userData = nullptr, bool codesOnly_InternalUse = false) noexcept;

		/// <summary>
		/// Returns the function called when a progress log message is written, or nullptr if none has been registered.
		/// </summary>
		/// <returns></returns>
		LogFunc* getProgressLogger() const noexcept;

		///@}


		///@{
		///@name Getting Output Images

		/// <summary>
		/// Return the 32-bit image resulting from the merge of the input images, or conversion of single image to pseudo 32-bit image.
		/// </summary>
		/// Performs alignment (if requested), merging, ghost removal (if requested) and noise reduction (if requested), if these have not yet been done.
		/// Note that this is automatically called if newcessary when processWithPreset() is called, if it has not been done previously.
		/// This method is provided separately to allow generation of 32-bit image without further processing, or saving of 32-bit image prior to processing.
		/// The lifetime of the returned HDRImage pointer is the lifetime of the PhotomatixEngine object, and it will become invalid after this object is destructed.
		Image* getHDRImage();

		/// <summary>
		/// Process the loaded images with the given preset.  If required, and not done yet, the images are loaded, aligned and merged first.
		/// </summary>
		/// <param name="preset">The preset with the processing settings</param>
		/// <param name="isPanorama">Some methods have the ability to process both ends of a pamorama in the same way so that they join up seamlessly.
		///							Set this to true if you are processing a 360 degree strip panorama.</param>
		/// <returns>A new image with the result of processing with that preset.</returns>
		Image processWithPreset(HDRPreset preset, bool isPanorama = false);

		//@}

		///@{
		///@name Accessing Input Images

		class ConstImageList;
		/// <summary>
		/// Returns an iterable object enumerating the images, which can be used in a range-based for loop.
		/// </summary>
		ConstImageList getImages() const noexcept;

		class ImageList;
		/// <summary>
		/// Returns an iterable object enumerating the images, which can be used in a range-based for loop.
		/// </summary>
		ImageList getImages() noexcept;

		///@}


		///@{
		///@name Less Used Options for Loading and Merging

		/// <summary>
		/// Bit depth to use for creating LDR output files.  Must be 8 or 16.
		/// </summary>
		/// If not set, bit depth defaults to the bit depth
		/// of the source images.  You would normally set this only if you were loading 8-bit images and wanted to save the merged
		/// result as a 16-bit image.  This must be set before calling processWithPreset().
		void setOutputBitDepth(int depth) noexcept;

		/// <summary>
		/// Bit depth to use for creating LDR output files.
		/// </summary>
		int getOutputBitDepth() const noexcept;



		/// <summary>
		/// Maximum number of parallel threads to use when processing.  This defaults to the number of cores in the system.
		/// </summary>
		/// Note that some operations use a single thread regardless of this setting.
		/// <param name="numThreads">Maximum number of threads to use.</param>
		void setNumThreads(int numThreads) noexcept;

		/// <summary>
		/// Maximum number of parallel threads to use when processing.
		/// </summary>
		int getNumThreads() const noexcept;

		/// <summary>
		/// Set the default color temperature that will be used when loading RAW images.
		/// </summary>
		/// This is ignored for any other type of image.  It must be called before image data is loaded.
		/// You can also call the method with the same name in individual Image objects to set colour temperature differently for each one.
		/// Color temperatures set on individual photos will override values set here.
		/// <param name="colorTemp">Color temperature to use</param>
		/// <param name="selection">Which photos to apply the color temperature to.
		///   If there is no metadata found for the image, or if there are images with duplicate metadata, no color temperature will be applied,
		///   unless it was set for EColorTempSelection_All, or the values for EColorTempSelection_OverExposed, EColorTempSelection_Normal and
		///   EColorTempSelection_UnderExposed are all the same.
		///   Supplying EColorTempSelection_All sets the color temperature for all photos, overriding any temperatures that were set for overexposed, underexposed
		///   or normal exposures.  You can set these subsequently to a call with EColorTempSelection_All however, to change the color temperature for just some
		///   sets of photos.</param>
		void setRAWColorTemperature(int colorTemp, EColorTempSelection selection = EColorTempSelection::All);

		/// <summary>
		/// The default color temperature that will be used when loading RAW images.
		/// </summary>
		int getRAWColorTemperature(EColorTempSelection selection = EColorTempSelection::All) const;


		/// <summary>
		/// Set the default color profile that will be used when loading RAW images.
		/// </summary>
		/// This is ignored for any other type of image.  It must be called before any images are added to the PhotomatixEngine object.
		/// <param name="profile">Color profile to use</param>
		void setRAWColorProfile(EColorProfile profile) noexcept;

		/// <summary>
		/// The default color profile that will be used when loading RAW images.
		/// </summary>
		EColorProfile getRAWColorProfile() const noexcept;

		/// <summary>
		/// Turn on/off automatic brightness adjustment when loading RAW images.
		/// </summary>
		/// When true, the image brightness is adjusted to preserve highlights when loading (default).
		/// When false, no adjustments are made, which can be useful when processing video frames which must have consistent brightness.
		/// This is ignored for any other type of image.  It must be set before image data is loaded.
		void setRAWAutoBrightness(bool autoBrightness) noexcept;

		/// <summary>
		/// Whether automatic brightness adjustment will be used by default when loading RAW images.
		/// </summary>
		bool getRAWAutoBrightness() const noexcept;

		/// <summary>
		/// Turn on/off highlight recovery when loading RAW images.
		/// </summary>
		/// If highlight recovery is turned on, images may be significantly darker.  It's recommended to turn this on
		/// only if processing a single image.
		/// This is ignored for any other type of image.  It must be called before image data is loaded.
		void setRAWHighlightRecovery(bool highlightRecovery) noexcept;

		/// <summary>
		/// Whether highlight recovery will be used by default when loading RAW images.
		/// </summary>
		bool getRAWHighlightRecovery() const noexcept;

		/// <summary>
		/// How to calculate the tonal curve when reading in RAW files.
		/// See ETonalCurveSource for valid values.
		/// </summary>
		void setTonalCurveSource(ETonalCurveSource tonalCurveSource) noexcept;

		/// <summary>
		/// How to calculate the tonal curve when reading in RAW files.
		/// See ETonalCurveSource for valid values.
		/// </summary>
		ETonalCurveSource getTonalCurveSource() const noexcept;



		/// <summary>
		/// If true, TIFF files as RAW.  This is for the Canon 1D and Phase One P45 cameras that produce RAW files with the .tif extension.
		/// Must be set before any files are added.
		/// </summary>
		void setTreatTiffAsRaw(bool treatTiffAsRaw) noexcept;

		/// <summary>
		/// Whether TIFF files are treated as RAW when loading.
		/// </summary>
		bool getTreatTiffAsRaw() const noexcept;


		/// \cond PRIVATE
		/// <summary>
		/// Internal use.
		/// </summary>
		bool getLogCodesOnly() const noexcept;

		/// <summary>
		/// Get the user data that will be passed for each log message.
		/// </summary>
		void *getLogUserData() const noexcept;
		/// \endcond


#ifdef ALIGN_INCLUDED
		/// <summary>
		/// Turn on/off cropping of images after alignment.  Defaults to cropping turned on.
		/// </summary>
		/// If cropping is off, the output image will be the same size as the input images, but there may be some garbled data around the edges of the final image.
		void setCropAfterAlign(bool crop) noexcept;
		
		/// <summary>
		/// Whether cropping after alignment is on or off.
		/// </summary>
		/// <returns></returns>
		bool getCropAfterAlign() const noexcept;
#endif

#ifdef DEGHOST_INCLUDED
		/// \cond PRIVATE
		/// <summary>
		/// Internal use
		/// </summary>
		/// <param name="maskFile"></param>
		/// <param name="maskWidth"></param>
		/// <param name="maskHeight"></param>
		/// <param name="maskNumImages"></param>
		void setDeghostingMaskFile(const char* maskFile, int maskWidth, int maskHeight, int maskNumImages);
		/// \endcond
#endif


		/// <summary>
		/// Manually supply the spacing, in EV units, between each supplied image in a bracketed set.
		/// </summary>
		/// This information is usually inferred from the metadata of the loaded images, but it can be overridden here, or supplied if exposure metadata is missing or invalid.
		void setCustomEVSpacing(float customEVSpacing) noexcept;
		
		/// <summary>
		/// Manually supplied EV spacing between images.
		/// </summary>
		/// <returns>EV spacing supplied, or 0 if the spacing is to be determined from the image metadata</returns>
		float getCustomEVSpacing() const noexcept;

		/// <summary>
		/// Set maximum megapixels of source images that the engine will load.  Set to 0 to allow any size image (default).
		/// </summary>
		void setMaxSourceMegapixels(float maxMegapixels) noexcept;
		
#ifdef SCRATCH_DISK_IMPLEMENTED
		// ----- Performance Options -----
		// Use scratch disk for Details Enhancer and Fusion/Natural

		/// <summary>
		/// If true, use temporary files to hold intermediate results when processing files using the Details Enhancer and Fusion/Natural algorithms.
		/// This will reduce performance, but also use significantly less memory.  Use only if you are running out of memory when processing large images with these algorithms.
		/// </summary>
		void setUseScratchDisk(bool useScratchDisk);

		/// <summary>
		/// Whether temporary files will be used to hold intermediate results for the Details Enhancer and Fusion/Natural algorithms to save memory.
		/// </summary>
		bool getUseScratchDisk() const;
#endif

		/// <summary>
		/// Directory to use for saving temporary files if needed.  Defaults to system default temporary directory.
		/// </summary>
		/// <param name="scratchDirectory"></param>
		void setScratchDirectory(const char* scratchDirectory);

		/// <summary>
		/// Directory to use for saving temporary files if needed.
		/// </summary>
		const char *getScratchDirectory() const noexcept;

#ifdef INTERNALFUNC_INCLUDED
		/// \cond PRIVATE
		/// <summary>
		/// Internal use
		/// </summary>
		void setLastProcess(bool lastProcess) noexcept;
#endif

#ifdef FUSION_INCLUDED
		/// <summary>
		/// Changes to Detail Contrast setting in Fusion/Natural relating to fixing moire artifacts when using this algorithm for freeze-frame movies.
		/// Values range from 1-7 for different mitigation strategies.
		/// </summary>
		void setMoireFix(int moireFix) noexcept;

		/// <summary>
		/// Returns the moire pattern fixing algorithm in use for the Fusion/Natural method.
		/// </summary>
		int getMoireFix() const noexcept;
		/// \endcond
#endif

		/// @}

		///@{
		///@name Other Actions

		/// <summary>
		/// Load the images, perform noise reduction if requested, and move the image data into the PhotomatixEngine object ready for alignment and HDR generation.
		/// </summary>
		void loadImages();
	
#ifdef ALIGN_INCLUDED	
		/// <summary>
		/// Align the images so that they are ready for HDR generation, using the settings supplied by setAlignmentType(), setAlignmentMaxOffset(), setCropAfterAlign().
		/// </summary>
		/// This is called automatically during HDR generation if it has not already been called.
		/// It is provided as a separate method mainly to allow saving of aligned images without needing to do further processing.
		void alignImages();
#endif

		///@}


		/// <summary>
		/// Iterable collection of images in a PhotomatixEngine object, usually obtained by calling PhotomatixEngine::getImages() rather than being directly instantiated.
		/// </summary>
		class EXPORT ImageList
		{
		public:
			ImageList(const PhotomatixEngine& engine) noexcept : photomatixEngineRef(engine) { }
			/// <summary>
			/// Returns start iterator
			/// </summary>
			const Image *begin() const noexcept;
			/// <summary>
			/// Returns start iterator
			/// </summary>
			Image* begin() noexcept;
			/// <summary>
			/// Returns end iterator
			/// </summary>
			const Image *end() const noexcept;
			/// <summary>
			/// Returns end iterator
			/// </summary>
			Image* end() noexcept;
			using iterator = Image*;
			using const_iterator = const Image*;
			/// <summary>
			/// Image with the specified index
			/// </summary>
			const Image& operator[](size_t idx) const;
			/// <summary>
			/// Image with the specified index
			/// </summary>
			Image& operator[](size_t idx);
			/// <summary>
			/// Number of images in the PhotomatixEngine object
			/// </summary>
			/// <returns></returns>
			size_t size() const noexcept;
		private:
			const PhotomatixEngine& photomatixEngineRef;
		};

		/// <summary>
		/// Iterable collection of const images in a PhotomatixEngine object, usually obtained by calling PhotomatixEngine::getImages() rather than being directly instantiated.
		/// </summary>
		class EXPORT ConstImageList
		{
		public:
			ConstImageList(const PhotomatixEngine& engine) noexcept : photomatixEngineRef(engine) { }
			/// <summary>
			/// Returns start iterator
			/// </summary>
			const Image* begin() const noexcept;
			/// <summary>
			/// Returns end iterator
			/// </summary>
			const Image* end() const noexcept;
			using iterator = const Image*;
			using const_iterator = const Image*;
			/// <summary>
			/// Image with the specified index
			/// </summary>
			const Image& operator[](size_t idx) const;
			/// <summary>
			/// Number of images in the PhotomatixEngine object
			/// </summary>
			size_t size() const noexcept;
		private:
			const PhotomatixEngine& photomatixEngineRef;
		};


		/// \cond PRIVATE
		~PhotomatixEngine() noexcept;
		PhotomatixEngine(const PhotomatixEngine&);
		PhotomatixEngine(PhotomatixEngine&&) noexcept;
		PhotomatixEngine& operator=(const PhotomatixEngine&);
		PhotomatixEngine& operator=(PhotomatixEngine&&) noexcept;
#ifdef CA_REDUCTION_INCLUDED
		void setCAReductionEnabled(bool caReductionEnabled) noexcept;
		bool getCAReductionEnabled() const noexcept;
#endif
		/// \endcond

	private:
		struct PhotomatixEngineInstanceData;
		PhotomatixEngineInstanceData* instanceData = nullptr;
		void checkImageCompatibility(const Image &img);
		bool hasLoadedImages() const noexcept;
		//void alignCallback(float prog, int* cancel);
		//void deghost();
		//Image average(const HDRPreset& preset);
		//Image average(const HDRPreset& preset);
		//Image average(const HDRPreset& preset);
		//Image average(const HDRPreset& preset);
		//Image average(const HDRPreset& preset);
		friend void alignCallback(float prog, int* cancel, void* data);
	};

	/// \cond PRIVATE
	const char* getLibPath() noexcept;
	/// \endcond

}  // Namespace

#endif
